(function (window, document) {

    const states = {
        closed: "closed",
        grid: "grid",
        vars: "vars",
        retargeting : "retargeting"
    };
    var config = {
        dcl: false, // DOMContentLoaded
        state: states.closed,
        defaultHeights: {  // defaultHeight
            grid: 355,
            vars: 375,
            retargeting: 375,
            closed: 0
        },
        lastSeenView: null,
        dataCollectionOption: 1
    };
    var isVarsPageReady = false; //this is state, not config, so don't put it in the config. It means the variables.controller is alive and listening for postMessages
    var isRetargetingPageReady = false; //this is state, not config, so don't put it in the config. It means the retargeting.controller is alive and listening for postMessages
    var gotVariablesFromServer = false;
    var advertiserData = {};   //used for vars page {name: "", id: nnn}
    var chosenDataLayerName = null;
    var pageHasAOneTag = false; //vars page needs to know if we're expecting a response from server, so we'll know to wait for that response before pm'ing the vars over to the page.
    var pubsub = {};
    var forcedInitHandler = null;

    var OTT = {

        dictionary : {},

        beforeInit: function () {
            // check blocklist - for some sites we shouldnt add the content script and continue the flow.
            // notice setExtensionFlag wont occour meaning the extension wont be active for this tabs
            if (this._checkBlockList() || this._shouldAbort()) {
                return;
            }

            // Listen to messages from background script (using port communication)
            // and listen to postmessages of active iframe control

            this._flagSetSuccessfuly = this._setExtensionFlag();

            if(this._flagSetSuccessfuly) {
                this._registerEvents();

                // Track conversion tags
                this._trackOldTags();
            }
        },

        _checkBlockList: (function () {
            var list = ["support.sizmek.com", "mdx.sizmek.com", "www.google.com", "www.google.co.il", "www.facebook.com", "calendar.google.com"];
            return function () {
                var domain = OTTInfra.extractDomain();
                if (~list.indexOf(domain)) {
                    return true;
                }
                return false;
            };
        })(),

        _shouldAbort: function() {
            return OTTInfra.isPdf()
        },

        _setExtensionFlag: function () {
            // Flag indicating tool is active
            try {
                var ott = document.createElement("iframe");
                ott.style.display = "none";
                ott.name = "ott_flag";
                // Why html? because thats all we have at this stage
                var parent = document.getElementsByTagName("html");
                if(parent.length) {
                    parent[0].appendChild(ott);
                    return true;
                }
            } catch(e) { }
            return false;
        },
        getActiveTab: function(callback) {
            chrome.tabs.query({ lastFocusedWindow: true, active: true }, function (tabs) {
                callback(tabs[0]);
            });
        },
        _registerEvents: function () {
            var $this = this;
            // handling messages received from background.js:
            this.port = chrome.runtime.connect();
            this.port.onMessage.addListener(function (msg) {
                if (msg && msg.command) {
                    $this[msg.command].apply($this, [msg]);
                }
            });

            chrome.runtime.onMessage.addListener(function(msg, sender, sendResponse) {
                if (msg && msg.command) {
                    $this[msg.command].apply($this, [msg]);
                }

                sendResponse({status:true})
                return true;
            });

            this._listenToPageCommands();
            // Create grid footer on DOMContentLoaded
            document.addEventListener('DOMContentLoaded', function () {
                config.dcl = true;
                $this.createGridFooter();

                // Responses from BackgroundService
                OTTInfra.pm.bind("BackgroundServiceResponse", function(data, msg) {
                    $this[data.command].apply($this, [data,msg]);
                }, $this, $this._grid.contentWindow);
            }, false);

            this._listenToAppCommands();

        },
        setTabId: function(msg) {
            // After content.js is connected with background.js, it sets its tabId to be used when storing data
            this.tabId = msg.data.tabId;
        },
        oneTagFound: function(msg) {
            // Some actions can only happen in backgroud.service. To run commands in background.js
            // we use this.port
            this.port.postMessage({ tabId: this.tabId, command: "oneTagFound", data: {} });
        },

        _listenToPageCommands: function () {
            var $this = this;
            // Listen to postmessages from window (OneTag.js)
            addEventListener("message", function (e) {
                // Accept messages from window only
                if (e.srcElement.location.href !== window.location.href) { return; }
                try {
                    var msgData = e.data;
                    if (msgData.key == "OTT" && msgData.command) {
                        $this.handleCommand(msgData.command, msgData, e);
                    }
                } catch (e) { }
            }, false);

            // Command handlers subscription
            /*this.addCommandListener(OTTEnums.commands.addRows, function (data) {
                this.send(OTTEnums.commands.addRows, { category: data.category, data: data.data });
            });

            this.addCommandListener(OTTEnums.commands.legacy, function (data) {
                var data = data.data; // actual tag data { qs:..., receiveTime:..., sendTime:... }
                if (!data || !data.qs || !data.qs.cn) {
                    return;
                }

                // Add to conversion/retargeting grid
                switch (data.qs.cn) {
                    case "as": {
                        this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.conversion, data: { id: data.qs.activityid, triggeringRuleIds: "-" } });
                        break;
                    }
                    case "dt": {
                        this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.retargeting, data: { id: data.qs.tid, setValue: data.qs.tval, triggeringRuleIds: "-" } });
                        break;
                    }
                    case "ntt": {
                        this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.retargeting, data: { id: data.qs.tid, setValue: data.qs.tval, triggeringRuleIds: "-" } });
                        break;
                    }
                }

                config.oldTagFound = config.oldTagFound || 0;
                config.oldTagFound++;

                if (config.oldTagFound && config.oneTagFound) {
                    // TODO: make this message pop in a modal (NEXT PHASE?)
                    console.log("We love Rum & Coke, but not all mixtures work as well. The page includes a mixture of both conversion/retargeting tag code and VersaTag code. This might cause our tags to not work as expected. We recommend that you remove the conversion/retargeting tags and use only the VersaTag code.")
                }

                // Add timeline row
                this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.timelines, data: { id: data.qs.activityid || data.qs.tid, receiveTime: data.receiveTime, sendTime: data.sendTime } });
            });

            this.addCommandListener(OTTEnums.commands.oneTagFound, function (data) {
                // Change tab icon to let the user know a tag has been found here
                // (we need the tab id so it will be done from background.js
                this.send(OTTEnums.commands.oneTagFound, { data: {} });
                pageHasAOneTag = true;
            });

            this.addCommandListener(OTTEnums.commands.parseServerResponse, function (data) {
                var rows = this._parseServerResponse(data);

                // mark a OneTag found on page ()
                config.oneTagFound = config.oneTagFound || 0;
                config.oneTagFound++;

                this._validateServerResponse(data, rows);

                for (var group in rows) {
                    this.send(OTTEnums.commands.addRows, { category: group, data: rows[group] });
                }

                //now that we have the oneTag ID, use it to get the advertiser ID.
                this._getAdvertiserID(data.data.tagManagerId);
            });

            this.addCommandListener(OTTEnums.commands.requestDataCollectionOption, function (ignore, event) {
                if (config.isInitialized) { // if not, OneTag.js will ask us again in a few milisecs, so no further handling is needed
                    // Dont be mistaken! this message goes to OneTag.js, not background.js. When the user sets the data collection option
                    // we fire a message to backgroud.js to set the general config with the selected option, so future calls will have the right option
                    event.source.postMessage({ key: "OTT", command: OTTEnums.commands.setDataCollectionOption, data: { dataCollectionOption: config.dataCollectionOption } }, "*");
                }
            });*/

            this.addCommandListener(OTTEnums.commands.pageVariablesResult, function (data) {
                //we got a result back from the page
                //write the result to 'dictionary'
                this._setVariablesDictionary(data.data);

                //do we need to tell background about this?

                //decide what to show next.
                this.showPageVariables();
            });

            this.addCommandListener(OTTEnums.commands.legacy, function(data) {
                $this.legacy(data.data);
            });
        },

        legacy: function(data) {
            // actual tag data { qs:..., receiveTime:..., sendTime:... }
            if (!data || !data.qs || !data.qs.cn) {
                return;
            }

            var tagTypeSupported = false;
            // Add to conversion/retargeting grid
            switch (data.qs.cn) {
                case "as": {
                    this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.conversion, data: { id: data.qs.activityid, triggeringRuleIds: "-" } });
                    tagTypeSupported = true;
                    break;
                }
                case "dt": {
                    this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.retargeting, data: { id: data.qs.tid, setValue: data.qs.tval, triggeringRuleIds: "-" } });
                    tagTypeSupported = true;
                    break;
                }
                case "ntt": {
                    this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.retargeting, data: { id: data.qs.tid, setValue: data.qs.tval, triggeringRuleIds: "-" } });
                    tagTypeSupported = true;
                    break;
                }
            }

            if(!tagTypeSupported) {
                return;
            }

            config.oldTagFound = config.oldTagFound || 0;
            config.oldTagFound++;

            if (config.oldTagFound && config.oneTagFound) {
                // TODO: make this message pop in a modal (NEXT PHASE?)
                console.log("We love Rum & Coke, but not all mixtures work as well. The page includes a mixture of both conversion/retargeting tag code and VersaTag code. This might cause our tags to not work as expected. We recommend that you remove the conversion/retargeting tags and use only the VersaTag code.")
            }

            // If we reached here, legacy tag (cn=as) was found in this page, therefore we should init extension to show those rows
            if(!forcedInitHandler) {
                var $this = this;
                forcedInitHandler = setTimeout(function() {
                    // If extension is still not initialized, this page probably contains only cn=as tags
                    // ebOneTag.js handshake is missing
                    if(!config.isInitialized) {
                        $this.init({config: {}});
                    }
                }, 5000);
            }


            // Add timeline row
            this.send(OTTEnums.commands.addRows, { category: OTTEnums.category.timelines, data: { id: data.qs.activityid || data.qs.tid, receiveTime: data.receiveTime, sendTime: data.sendTime } });
        },

        handleCommand: function (command, data, event) {
            var record = pubsub[command];

            if (record) {
                for (var i = 0; i < record.length; i++) {
                    record[i].apply(this, [data, event]);
                }
            }
        },

        addCommandListener: function (command, handler) {
            pubsub[command] = pubsub[command] || [];
            pubsub[command].push(handler);
        },

        _parseServerResponse: function (data) {
            var res = {};
            res["3rdparty"] = [];
            res["conversion"] = [];
            res["retargeting"] = [];
            res["timelines"] = [];

            var data = data.data;
            var receiveTime = data.receiveTime;
            var sendTime = data.sendTime;
            var tagManagerId = data.tagManagerId;
            var arr; var record;

            for (var ruleId in data) {
                arr = data[ruleId]["3rdPartyActivities"];
                if (arr) {
                    for (var i = 0; i < arr.length; i++) {
                        record = arr[i];
                        res["3rdparty"].push({
                            id: record["Id"],
                            name: "",
                            triggeringRuleIds: "",
                            _triggeringRuleIds: [ruleId],
                            _triggeringRuleNames: "",
                            code: record["Urls"] || record["Urls"],
                            tagManagerId: tagManagerId
                        });
                    }
                }

                arr = data[ruleId]["ConversionActivities"];
                if (arr) {
                    for (var i = 0; i < arr.length; i++) {
                        record = arr[i];
                        res["conversion"].push({
                            id: record["Id"],
                            name: "",
                            triggeringRuleIds: "",
                            _triggeringRuleIds: [ruleId],
                            _triggeringRuleNames: "",
                            type: this._enumTypeToTerm(record["Type"]),
                            parameters: record["Params"], // TODO: check with Naor why data is missing (maybe not yet implemented)
                            thirdPartyTags: record["ThirdParties"],
                            pathToConversion: record["PathToConversion"],
                            countingMethod: record["CountingMethod"],
                            tagManagerId: tagManagerId,
                            status: "200 OK"
                        });

                        OTTInfra.addDistinct(res["timelines"], {
                            id: record["Id"],
                            name: "Conversion", // as in tag type (conversion/retargeting)
                            triggeringRuleIds: [ruleId],
                            type: this._enumTypeToTerm(record["Type"]),
                            sendTime: sendTime,
                            receiveTime: receiveTime
                        });
                    }
                }

                arr = data[ruleId]["RetargetingActivities"];
                if (arr) {
                    for (var i = 0; i < arr.length; i++) {
                        record = arr[i];


                        // Values returened contains all values for all entries of retargeting
                        // this is why we need to filter out the relevant data for this specific retargeting record
                        var setValue = "";
                        if(record["Values"]) {
                            var regex = new RegExp(record["Id"] + "=(.*?)(&|$)");
                            var match = regex.exec(record["Values"])
                            if(match.length > 0) {
                                setValue = match[1];
                            }
                        }

                        res["retargeting"].push({
                            id: record["Id"],
                            name: "",
                            triggeringRuleIds: "",
                            _triggeringRuleIds: [ruleId],
                            _triggeringRuleNames: "",
                            setValue: setValue,
                            status: "200 OK",
                            tagManagerId: tagManagerId
                        });

                        OTTInfra.addDistinct(res["timelines"], {
                            id: record["Id"],
                            name: "Retargeting", // as in tag type (conversion/retargeting)
                            triggeringRuleIds: [ruleId],
                            type: "Retargeting Request",
                            sendTime: sendTime,
                            receiveTime: receiveTime
                        });
                    }
                }
            }

            // ready for addRows
            return res;
        },

        _enumTypeToTerm: function (type) {
            if ((type * 1) == 1) {
                return "Counter";
            }
            return "Sales";
        },

        _validateServerResponse: function (rows) {
            // TODO: vaildate data (look for two rows with the same id, but different ruleId - meaning two rules
            // are firing the same activity)


            // this method operate in two levels:
            // 1. rearange / analyze data to understand if data impaired
            // 2. create a result object explaining all issues with current data
            var counter = {};
            var record, id, validation = [];
            for (var category in rows) {
                record = rows[category];
                id = record.id
                if (id && record.ruleIds) {
                    // create a new record for each row id
                    // and counting the number of ruleId
                    counter[id] = counter[id] || {};
                    counter[id].ruleIds = counter[id].ruleIds || [];
                    counter[id].ruleIds.push(record.ruleIds);
                }
            }

            for (var id in counter) {
                if (counter[id].ruleIds.length > 1) {
                    validation.push({
                        msg: "Activity/Retargeting of id " + id + " has been fired by two different rules at the same time",
                        id: id
                    });
                }
            }

            // New tag is in this page for sure,
            // but if there is an old tag we should show the following:
            if (config.oldTagFound && config.oneTagFound) {
                validation.push({
                    msg: "We love Rum & Coke, but not all mixtures work as well. The page includes a mixture of both conversion/retargeting tag code and VersaTag code. This might cause our tags to not work as expected. We recommend that you remove the conversion/retargeting tags and use only the VersaTag code."
                });
            }

            // display messages
            for (var i = 0; i < validation.length; i++) {
                // TODO: fire an event to display message in modal
                // IMPOTANT: make event handler smart to distinct multiple messages (display message once per case)
                // cause old tag hanlder fire the same messages, depends on tag firing timing

                console.log(validation[i].msg);
            }
        },

        _listenToAppCommands: function () {

            var $this = this;

            // handling messages received from grid/vars control:

            // Wait till grid is ready (iframe loaded) before requesting initialization
            // (which might update grid values)
            // sending authenticationToken to background.js
            var gridReadyCount = 0;
            OTTInfra.pm.bind(OTTEnums.commands.gridReady, function (data) {
                data = data || {};
                data.tabId = $this.tabId;
                if(gridReadyCount == 0) {
                    data.pageRefreshed = (performance.navigation.type == 1);
                } else {
                    data.pageRefreshed = false;
                }

                // Count number of times gridReady was called
                gridReadyCount++;
                config.lastSeenView = data.currentView;

                $this.send(OTTEnums.commands.requestInit, data);
            });

            OTTInfra.pm.bind(OTTEnums.commands.enrichDataRowsAfterLogin, function (data) {
                //{ sessionId: AuthenticationService.getCachedAuthenticationToken() }
                data = data || {};

                // This function clears data rows in UI only
                // enrichment will then go through each data row and enrich their values by querying our platform
                // once thats done, each row enriched (along any other row which didnt required any enrichment) will be
                // sent to UI via addRows
                $this.clearUIRows();
                $this.send(OTTEnums.commands.enrichDataRowsAfterLogin, data);
            });

            OTTInfra.pm.bind(OTTEnums.commands.variablesReady, function (data) {
                if (data.value === false) {
                    //he ded.
                    isVarsPageReady = false;
                }
                else {
                    isVarsPageReady = true;
                    $this.showPageVariables();
                }
            });

            // Expands/collapse grid
            OTTInfra.pm.bind(OTTEnums.commands.toggleGrid, function (msg) {
                $this.toggleGrid(msg.sender);
            });

            OTTInfra.pm.bind(OTTEnums.commands.toggleVars, function (msg) {
                $this.toggleVars(msg.sender);
            });

            OTTInfra.pm.bind(OTTEnums.commands.toggleRetarget, function (msg) {
                $this.toggleRetarget(msg.sender);
            });

            OTTInfra.pm.bind(OTTEnums.commands.close, function (msg) {
                $this.close();
            });

            // set view to a different height
            // usualy done when a modal is openned or closed
            OTTInfra.pm.bind(OTTEnums.commands.setViewHeight, function (msg) {
                $this.setViewHeight(msg.height);
            });

            // set view to a different height
            // usualy done when a modal is openned or closed
            OTTInfra.pm.bind(OTTEnums.commands.setDataCollectionOption, function (msg) {
                // We do not need it anymore, message came from background js
                //$this.setDataCollectionOption(msg);
                // Reload page after setting the new collection option
                setTimeout(function () {
                    location.reload();
                }, 300);
            });

            OTTInfra.pm.bind(OTTEnums.commands.requestPageVariables, function (data) {
                //for sure the page is ready if we got this
                isVarsPageReady = true;
                //trigger the scan, passing along any specific request
                $this.analyzePageVariables(data);
            });

            /*OTTInfra.pm.bind(OTTEnums.commands.writeAdvertiserVariables, function (data) {
                //for sure the page is ready if we got this
                isVarsPageReady = true;
                var variablesData = $this._prepVariablesForServer(data);
                //tell backgroundjs to send it
                $this.send(OTTEnums.commands.writeAdvertiserVariables, {sessionId: data.sessionId, data: variablesData});
            });*/

        },

        init: function (msg) {
            //this happens when the contentjs has done handshake with backgroundjs
            OTTInfra.mergeObj(msg.config, config, true);

            if (config.state !== states.closed) {
                this._show();
            }

            config.isInitialized = true;

            if (config.state === states.grid) {
                //tell the grid about which colectionOption was in storage
                let mode;
                switch (config.dataCollectionOption) {
                    case 0:
                        mode = 'Full Data';
                        break;
                    case 1:
                        mode = 'None';
                        break;
                    case 2:
                        mode = 'Partial Data';
                        break;
                }

                if (mode) {
                    //this.send(OTTEnums.commands.showDataCollectionOption, {selectionMode: mode});//  {selectionMode: 'None' / 'Partial Data' / 'Full Data'}
                    OTTInfra.pm({
                        target: this._grid.contentWindow,
                        type: OTTEnums.commands.showDataCollectionOption,
                        data: { selectionMode: mode }
                    });
                }
            }

            // Send data rows to grid
            // TODO: no need, it should be done in BackgroundService
            //this.addRows(msg);

            // Fire pending rows
            if (config.pendingMessages) {
                var pending = config.pendingMessages;
                for (var i = 0; i < pending.length; i++) {
                    this.send(pending[i].command, pending[i].data);
                }
            }

            //grid is the default for some reason, so if we opened in vars view, switch to it now.
            if (config.state == states.vars) {
                //tell ng switch to vars view now
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.setDefaultView,
                    data: { view: 'variable' }
                });
                //and trigger the scan
                this.analyzePageVariables();
            }

             //grid is the default for some reason, so if we opened in vars view, switch to it now.
             if (config.state == states.retargeting) {
                //tell ng switch to vars view now
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.setDefaultView,
                    data: { view: 'retargeting' }
                });
                //and trigger the scan
               // this.analyzePageRtargeting();
            }
        },

        // Send data to backgroud
        send: function (command, data) {
            if (config.isInitialized || command == OTTEnums.commands.requestInit || command == OTTEnums.commands.writeAdvertiserVariables || command == OTTEnums.commands.oneTagFound || command == OTTEnums.commands.getAdvertiserId || command == OTTEnums.commands.serverVariables || command == OTTEnums.commands.toggleGrid) {
                if(data) {
                    data.command = command;
                } else {
                    data = { command: command };
                }

                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: command,
                    data: data
                });
            } else {
                config.pendingMessages = config.pendingMessages || [];
                config.pendingMessages.push({ command: command, data: data || {} });
            }
        },

        _trackOldTags: function () {
            this.inject([this.scriptFromFile("legacy.js")]);
        },

        scriptFromFile: function (file) {
            var script = document.createElement("script");
            script.src = chrome.runtime.getURL(file);
            return script;
        },

        scriptFromSource: function (source) {
            var script = document.createElement("script");
            script.textContent = source;
            return script;
        },

        inject: function (scripts) {
            if (scripts.length === 0)
                return;
            var otherScripts = scripts.slice(1);
            var script = scripts[0];
            var $this = this;
            var onload = function () {
                script.parentNode.removeChild(script);
                $this.inject(otherScripts);
            };

            if (script.src) {
                script.onload = onload;
                (document.head || document.getElementsByTagName("html")[0]).appendChild(script);
            } else {
                (document.head || document.getElementsByTagName("html")[0]).appendChild(script);
                onload();
            }
        },

        addRows: function (msg) {
            // Data example:
            // msg.data = { timelines: [row1, row2...], conversion: [ row1, row2... ] }
            var args = arguments;

            if (args.length == 2) {
                var res = {};
                res[args[0]] = OTTInfra.isArray(args[1]) ? args[1] : [args[1]];

                msg = {};
                msg.data = res;
            }

            this.saveMessageData(msg.data);

            OTTInfra.pm({
                target: this._grid.contentWindow,
                type: OTTEnums.commands.addRows,
                data: msg.data
            });
        },

        clearUIRows: function() {
            OTTInfra.pm({
                target: this._grid.contentWindow,
                type: OTTEnums.commands.clearRows,
                data: {}
            });
        },

        saveMessageData: function(data) {
            this.currentGridData = this.currentGridData || {};
            for(var category in data) {
                this.currentGridData[category] = this.currentGridData[category] || [];
                OTTInfra.addDistinct(this.currentGridData[category], data[category]);
            }
        },

        writeAdvertiserVarsResult: function (msg) {
            //msg came from background
            //figure out what happened, and tell variables.controller about it.
            let result;
            try {
                result = JSON.parse(msg.data);
            }
            catch(e) {
                //copy the structure that server sends us
                result = {
                    error: {
                        errors: [
                            { innerMessage: "OTT writeAdvertiserVarsResult, unable to parse JSON response from server" }
                        ]
                    }
                };
            }

            //Tell controller
            OTTInfra.pm({
                target: this._grid.contentWindow,
                type: OTTEnums.commands.writeAdvertiserVarsResult,
                data: result
            });

            //tell the UI that anything that was created on server now has "exist on server' status.
            this.serverVariables( {data: result} );
        },

        close: function () {
            this._hide();
        },

        toggleVars: function (requestedBy) {
            //if toggleVars was requested by grid, ignore - the states don't allow you to swap - swapping is via the popup.
            if (requestedBy == states.grid || requestedBy == states.retargeting) {
                return;
            }

            //if vars is closed (grid is open or both are closed)
            //open vars
            if (config.state !== states.vars) {
                //it's possible the iframe doesn't exist yet...
                if (this._grid && this._grid.contentWindow) {
                this._show(states.vars);

                //tell angular to show vars view
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.setDefaultView,
                    data: { view: 'variable' }
                });

                //trigger the scan
                this.analyzePageVariables();
            }
            else {
                    var $this = this;
                    setTimeout(function() { $this.toggleVars.apply($this,[requestedBy]);}, 100);
                }
            }
            else {
                //it was alread vars, so close
                this._hide();
            }

            config.lastSeenView = states.vars;
        },

        toggleRetarget: function (requestedBy) {
            //if toggleRetarget was requested by grid, ignore - the states don't allow you to swap - swapping is via the popup.
            if (requestedBy == states.vars || requestedBy ==states.grid) {
                return;
            }

            //if Retarget is closed (grid or vars is open or both are closed)
            //open Retarget
            if (config.state !== states.retargeting) {
                //it's possible the iframe doesn't exist yet...
                if (this._grid && this._grid.contentWindow) {
                    this._show(states.retargeting);

                    //tell angular to show Retarget view
                    OTTInfra.pm({
                        target: this._grid.contentWindow,
                        type: OTTEnums.commands.setDefaultView,
                        data: { view: 'retargeting' }
                    });

                    //trigger the scan
                // this.analyzePageRtargeting();
                }
                else {
                    var $this = this;
                    setTimeout(function() { $this.toggleRetarget.apply($this,[requestedBy]);}, 100);
                }
            }
            else {
                //it was alread toggleRetarget, so close
                this._hide();
            }
            config.lastSeenView = states.retargeting;
        },

        toggleGrid: function (requestedBy) {
            //if toggleVars was requested by grid, ignore - the states don't allow you to swap - swapping is via the popup.
            if (requestedBy == states.vars || requestedBy == states.retargeting) {
                return;
            }
            var $this = this;
            var prevState = config.state;

            //if grid is closed (vars or retargeting is open or both are closed)
            //open grid
            if (config.state !== states.grid) {
                //it's possible the iframe doesn't exist yet...
                if (this._grid && this._grid.contentWindow) {
                this._show(states.grid);

                //tell angular to show grid view
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.setDefaultView,
                    data: { view: 'grid' }
                });

                // Re-show grid data
                if(config.lastSeenView && config.lastSeenView != states.grid) {
                    setTimeout(function() {
                        OTTInfra.pm({
                            target: $this._grid.contentWindow,
                            type: OTTEnums.commands.addRows,
                            data: $this.currentGridData
                        });
                    }, 100);
                }

                setTimeout(function() {
                    let mode;
                    switch (config.dataCollectionOption) {
                        case 0:
                            mode = 'Full Data';
                            break;
                        case 1:
                            mode = 'None';
                            break;
                        case 2:
                            mode = 'Partial Data';
                            break;
                    }
                    if (mode) {
                        //this.send(OTTEnums.commands.showDataCollectionOption, {selectionMode: mode});//  {selectionMode: 'None' / 'Partial Data' / 'Full Data'}
                        OTTInfra.pm({
                            target: $this._grid.contentWindow,
                            type: OTTEnums.commands.showDataCollectionOption,
                            data: { selectionMode: mode }
                        });
                    }
                }, 100);
                }
                else {
                    setTimeout(function() { $this.toggleGrid.apply($this,[requestedBy]);}, 100);
                }
            }
            else {
                //it was alread grid, so close
                this._hide();
            }

            config.lastSeenView = states.grid;
        },

        _show: function (state) {
            config.state = state || config.state;
            this.setViewHeight();

            //tell background about it
            //this.send(OTTEnums.commands.toggleGrid, { state: config.state });
            OTTInfra.pm({
                target: this._grid.contentWindow,
                type: OTTEnums.commands.toggleGrid,
                data: { state: config.state }
            });
        },

        _hide: function () {
            config.state = states.closed;
            this.setViewHeight();

            //tell background about it
            //this.send(OTTEnums.commands.toggleGrid, { state: config.state });
            OTTInfra.pm({
                target: this._grid.contentWindow,
                type: OTTEnums.commands.toggleGrid,
                data: { state: config.state }
            });
        },

        setDataCollectionOption: function (msg, event) {
            config.dataCollectionOption = msg.option;
            // Let background.js know of the new option. it will store it in config, so any new page loaded will have the
            // selected option
            this.send(OTTEnums.commands.setDataCollectionOption, { dataCollectionOption: config.dataCollectionOption });

            // Let OneTag.js know, if its not too late for its calls
            event.source.postMessage({ key: "OTT", command: OTTEnums.commands.setDataCollectionOption, data: { dataCollectionOption: config.dataCollectionOption } }, "*");
        },

        getDataCollectionOption: function () {
            return (typeof config.dataCollectionOption !== "undefined") ? config.dataCollectionOption : 1;
        },

        setViewHeight: function (height) {
            //console.log("setViewHeight: ", height , config.state);
            height = typeof height != "undefined" ? height : config.defaultHeights[config.state];
            if (this._grid) {
                this._grid.style.height = height + "px";
            }
        },

        analyzePageVariables: function (msg) {
            //if this comes from popup/background we won't have anything interesting in msg.data
            //if it comes from VariablesController / angular it might have something in there, 'search for this specific var name'
            chosenDataLayerName = msg && msg.dataLayerName || null;
            var $this = this;
            var script = this.scriptFromFile("js/analyzer.js");
            if (chosenDataLayerName) {
                script.setAttribute('data-dataLayerName', chosenDataLayerName);
            }
            $this.inject([script]);
        },

        showPageVariables: function () {
            //dictionary should have been updated by now. Send it to the tool iframe.
            if (this.dictionary && this.dictionary.fullList && this.dictionary.fullList.length && isVarsPageReady) {
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.pageVariablesData,
                    data: this.dictionary
                });
                if (this.dictionary.dataLayers.length > 1) {
                    this.setViewHeight(590);
                }
                else {
                    this.setViewHeight();
                }
            }
            else {
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.pageVariablesData,
                    data: {view: 'loader'}
                });
            }
        },

        _setVariablesDictionary: function (fromPage, fromServer) {
            //Here we have three streams of data (from two sources - fromPage contains two lists)
            //The data comes in asyncly at different times.
            //The enricher tells us what already exists the tag as defined on the server. But in the case of DataLayer items, it doesn't give us the name of the DataLayer.
            //So we need to cross-reference it with the dataLayer list from the page.
            //Then we need to update the 'fullList' of vars from teh page with the information we learned from cross referencing.

            //keep dict in scope for when 'this' changes in forEach...
            var dict = this.dictionary;

            //make sure dict arrays exist and are the right type before me mess with them.
            dict.fullList = dict.fullList || [];
            dict.dataLayers = dict.dataLayers || [];
            dict.alreadyExist = dict.alreadyExist || [];

            //update dict with any new stuff from page scan
            if (fromPage) {
                if(fromPage.dataLayers && OTTInfra.isArray(fromPage.dataLayers)) {
                    //copy it to the dict if the dict doesn't already have it. It doesn't matter very much if there are duplicates, but let's be nice and avoid even a tiny memory leak.
                    //This way if another scan gets triggered without the DataLayerName being submitted, we won't forget about the DataLayerName that was previously submitted.
                    fromPage.dataLayers.forEach(function(newDLItem) {
                        OTTInfra.pushUnique(newDLItem, dict.dataLayers);
                    });
                    //if a dataLAyerName was specified, make sure we send ONLY that datalayerName to the UI
                    if (chosenDataLayerName) {
                        dict.dataLayers = dict.dataLayers.filter(item => item.variableName == chosenDataLayerName);
                    }
                }

                //just replace the fullList
                if (fromPage.fullList && OTTInfra.isArray(fromPage.fullList)) {
                    dict.fullList = fromPage.fullList;
                }
            }

            //update dict with any new stuff from tag/enricher
            if (fromServer) {
                fromServer.forEach(function(newItemFromServer) {
                    //enricher seems to send some duplicates. I think it's because the OneTag.js repeatedly sends the info until background.js replies and tells him to stop.
                    //We prefer not to have duplicates in our dictionary, so use pushUnique.
                    OTTInfra.pushUnique(newItemFromServer, dict.alreadyExist);
                });
            }

            //merge the cross-referenced data into the fullList
            if (dict.fullList && dict.alreadyExist && gotVariablesFromServer) {
                dict.alreadyExist.forEach(function(serverItem) {
                    if (serverItem.type == "Data Layer") {
                        if  (dict.dataLayers) {
                            //Server knows about a Data Layer variable, but we don't know what the DataLayer itself is called.
                            //We need to get the data layer name. We will use "duck-typing" to find it.
                            //If the fullList contains an item that looks like "[someDataLayerNameFromPageDataLayersList].[VarNameFromServer]" we will consider it to be a match.
                            var searchVarName = serverItem.nameOnPage;
                            dict.dataLayers.forEach(function(dataLayerItem) {
                                var dataLayerName = dataLayerItem.variableName;
                                var search = dataLayerName + "." + searchVarName;
                                //OK now we know what to search for, search the list for it.
                                dict.fullList.forEach(function(listItem) {
                                    if (listItem.type == "Data Layer" && listItem.variableName == search) {
                                        //found a match. Flag it.
                                        listItem.existOnServer = true;
                                    }
                                });
                            });
                        }
                    }
                    else if (serverItem.type == "Global Variable") {
                        dict.fullList.forEach(function(listItem) {
                            if (listItem.variableName == serverItem.nameOnPage) {
                                listItem.existOnServer = true;
                            }
                        });
                    }
                    else if (serverItem.type == "Querystring") {
                        dict.fullList.forEach(function(listItem) {
                            if (listItem.variableName == serverItem.nameOnPage) {
                                listItem.existOnServer = true;
                            }
                        });
                    }
                });
            }

            //now make sure the .alreadyExist is present
            dict.fullList.forEach(function(listItem) {
                //if (gotVariablesFromServer) {
                listItem.existOnServer = listItem.existOnServer ? true : false;
                //}
                //else {
                ////commented out because not yet supported by ui/controller
                //    listItem.existOnServer = "Requires Login";
                //}
            });
        },

        /*_prepVariablesForServer: function (data) {
            var variables = {entities: []};
            data.selectedVariables.forEach(function (elem) {
                let varType;
                switch(elem.type) {
                    case "Data Layer":
                        varType = "DataLayerVariable";
                        break;
                    case "Global Variable":
                        varType = "PageJSVariable";
                        break;
                    case "Query String":
                        varType = "URLQueryStringParam";
                        break;
                };

                variables.entities.push({
                    type: "TagManagerVariable",
                    name: elem.variableName,
                    advertiserId: data.tagManagerSelectedAdvertisers.id,
                    variableType: varType,
                    dataType: elem.dataType,
                    variableName: elem.variableName
                });
            });

            return variables;
        },*/

        /*_getAdvertiserID: function (tagID) {
            this.send(OTTEnums.commands.getAdvertiserId, {oneTagID: tagID});
        },

        //this is the handler for OTTEnums.commands.getAdvertiserId when it comes back from the background.js
        getAdvertiserId: function (msg) {
            advertiserData.id = msg.data.result.advertiserId;
            advertiserData.name = msg.data.result.name;

            //tell the vars controller about it, if its open
            if (isVarsPageReady && advertiserData.id) {
                OTTInfra.pm({
                    target: this._grid.contentWindow,
                    type: OTTEnums.commands.advertisersData,
                    data: {selectedAdvertiser: advertiserData}
                });
            }

            //now that we have the advertiserID we can get the variables of this adv from the server.
            if (advertiserData.id) {
                //if we failed to get the id, it's most likely because we're not logged in, so don't bother continuing with this flow just yet.
            this._getVariablesFromServer(advertiserData.id);
            }
        },

        _getVariablesFromServer: function (advertiserId) {
            this.send(OTTEnums.commands.serverVariables, {advertiserId: advertiserId });
        },*/

        //this is the handler for OTTEnums.commands.serverVariables when it comes back from the background.js
        serverVariables: function (msg) {
            //msg.data.result[0].variableName .variableType .dataType .name
            let variables = [];
            let varType = "";
            if (msg.data && msg.data.result && Array.isArray(msg.data.result)) {
                msg.data.result.forEach(function(item) {
                    //translate varType from server API-speak to ui speak
                    switch(item.variableType) {
                        case "URLQueryStringParam":
                            varType = "Querystring";
                            break;
                        case "DataLayerVariable":
                            varType = "Data Layer"
                            break;
                        case "PageJSVariable":
                            varType = "Global Variable"
                            break;
                        default:
                            varType = item.variableType;
                    }

                    variables.push ({
                        variableName: item.name,
                        type: varType,
                        nameOnPage: item.variableName

                    });
                });

                this._setVariablesDictionary(null, variables);

            }
            gotVariablesFromServer = true;
            this.showPageVariables();
        },

        createGridFooter: function () {
            var version = chrome.runtime.getManifest().version.split(".").join("_");
            var iframe = document.createElement("iframe");
            iframe.id = "szmk_ott";
            iframe.src = "https://"+ OTTInfra.getEnviroment()+"/BurstingScript/OneTagExtesionTool.html?version=" + version;
            iframe.style = "overflow:hidden; position: fixed; bottom: 0;left: 0;right: 0;width: 100%; height: 300px;background-color: lightblue;color: black;transition: all 0.5s ease 0s;min-height:0px";
            iframe.style.display = "inline";
            //iframe.sandbox = "allow-same-origin allow-scripts allow-popups allow-forms";

            this._grid = iframe;
            this._grid.style.height = config.defaultHeights[config.state] + "px";

            // We dont need to hide the grid when created as we used to
            // (when we do, we tell background we did it, and it cancels the auto-expand of the toolbar
            //this._hide();

            document.body.append(iframe);
        },

        // Track receives bsResponse data and create a data row accordingly
        track: function (bsResponseRecord, tagInfo) {
            var response = {};
            OTTInfra.mergeObj(OTTInfra.cloneObj(bsResponseRecord), response);
            OTTInfra.mergeObj(OTTInfra.cloneObj(tagInfo), response);
            this.send(OTTEnums.commands.addRows, response);
        }
    };

    OTT.beforeInit();
})(window, document);
