(function() {
  const getVarsFromPage = function() {
    var dataLayerName =
      document.currentScript &&
      document.currentScript.getAttribute("data-dataLayerName");
    var list = [];
    var dataLayers = [];
    const MAX_LEVEL = 5;

    //use Set instead of Array for blocklist because Set guarantees elements are unique.
    var blocklist = new Set([
      "webkitStorageInfo",
      "Modernizr",
      "yepnope",
      "versaTagObj",
      "EBGVT",
      "EBGUIP",
      "OTTInfra",
      "OTTEnums",
      "intName",
      "eventData",
      "target",
      "postMessage",
      "XMLHttpRequest",
      "srcElement",
      "currentTarget",
      "relatedTarget",
      "lastIndexOf",
      "tagName",
      "data",
      "text",
      "innerHTML",
      "onload",
      "onclick",
      "Counter",
      "Timer",
      "type",
      "creationTime",
      "name",
      "clearInterval",
      "setAttribute",
      "ownerDocument",
      "createElement",
      "getElementsByTagName",
      "top",
      "parseInt",
      "encodeURIComponent",
      "escape",
      "clearTimeout",
      "setTimeout",
      "setInterval",
      "appendChild",
      "isNaN",
      "ActiveXObject",
      "attachEvent",
      "addEventListener",
      "log",
      "call",
      "undefined",
      "RegExp",
      "getUTCSeconds",
      "getUTCFullYear",
      "getUTCMinutes",
      "getUTCHours",
      "getUTCDate",
      "getUTCMonth",
      "Boolean",
      "Number",
      "Double",
      "Date",
      "Logging",
      "apply",
      "bind",
      "unescape",
      "resPath",
      "currentEbO",
      "extensionHooks",
      "_adConfig",
      "ebPtcl",
      "ebBigS",
      "ebResourcePath",
      "sHost",
      "winP",
      "EB",
      "EBG",
      "String",
      "body",
      "cWin",
      "charAt",
      "charCodeAt",
      "concat",
      "ebO",
      "endsWith",
      "flags",
      "fromCharCode",
      "getTime",
      "hasOwnProperty",
      "href",
      "includes",
      "indexOf",
      "lastIndex",
      "lastIndexOf",
      "length",
      "localeCompare",
      "location",
      "match",
      "parent",
      "parentElement",
      "prototype",
      "repeat",
      "replace",
      "search",
      "slice",
      "split",
      "src",
      "startsWith",
      "style",
      "substr",
      "substring",
      "tWin",
      "test",
      "toLocaleLowerCase",
      "toLocaleUpperCase",
      "toLowerCase",
      "toString",
      "toUpperCase",
      "trim",
      "valueOf",
      "win",
      "copyWithin",
      "every",
      "fill",
      "filter",
      "find",
      "findIndex",
      "forEach",
      "isArray",
      "join",
      "map",
      "pop",
      "push",
      "reduce",
      "reduceRight",
      "reverse",
      "shift",
      "some",
      "sort",
      "splice",
      "unshift",
      "callee",
      "_CCs",
      "_bannerCC",
      "_completeProfile",
      "_defaultProfile",
      "IFrameCC",
      "xhr",
      "$",
      "pm",
      "attributes",
      "_logicalEvents",
      "adConfig",
      "gEBMainWindow",
      "JSON",
      "ffs",
      "_adConfig",
      "navigator",
      "ebOArr",
      "$sf",
      "Array",
      "EB",
      "Math",
      "Object",
      "arguments",
      "body",
      "cWin",
      "document",
      "ebO",
      "currentResponse",
      "currentEbO",
      "location",
      "mraid",
      "style",
      "tWin",
      "win",
      "window"
    ]);

    const onBlocklist = function(name) {
      return blocklist.has(name) && !isDataLayer(name);
    };

    const isDataLayer = function(objName, parentObj) {
      //if it matches the "dataLayerName" we were told to search for, then that's the only thing we care about
      if (dataLayerName) {
        return objName == dataLayerName;
      } else {
        //if we were not told to find a specific dataLayerName, then it could be any of these.
        return (
          [
            "dataLayer",
            "utag_data",
            "universal_variable",
            "tm_param",
            "SizmekDataLayer"
          ].includes(objName) &&
          isWindow(parentObj) &&
          parentObj[objName]
        );
      }
    };

    const updateBlocklist = function() {
      //add to the blocklist anything contained in an empty iframe window
      var sandboxIframe = document.body.appendChild(
        document.createElement("iframe")
      );
      for (let windowProp in window) {
        if (
          !onBlocklist(windowProp) &&
          typeof sandboxIframe.contentWindow[windowProp] !== "undefined"
        ) {
          blocklist.add(windowProp);
        }
      }
      document.body.removeChild(sandboxIframe);
    };

    const isPrimitive = function(thing) {
      return (
        typeof thing === "string" ||
        typeof thing === "number" ||
        typeof thing === "boolean"
      );
    };

    const isWindow = function(obj) {
      return obj && obj.self && obj.self == obj;
    };

    const safeToIterate = function(obj) {
      try {
        return (
          obj !== null && !(obj instanceof Window) && "hasOwnProperty" in obj
        );
      } catch (e) {
        return false;
      }
    };

    const getType = function(val) {
      //server allowed types are [String, Integer]
      if (typeof val === "string") {
        return "String";
      } else if (typeof val === "number") {
        if (val == Math.round(val)) {
          return "Integer";
        } else {
          //use 'String' because rounding would be even worse.
          return "String";
        }
      } else if (typeof val === "boolean") {
        //use 'String' so platform can do if var.contains('true') {...
        return "String";
      }
    };

    function grabVariables(prop, str, level, isDataLayerFlag, callback) {
      semaphore--;
      if (!prop || level >= MAX_LEVEL) {
        if (semaphore == 0) {
          callback.apply(this);
        }
        return;
      }

      for (var p in prop) {
        if (prop.hasOwnProperty(p)) {
          var newStr = str ? str + "." + p : p;
          if (!onBlocklist(p)) {
            if (typeof prop[p] == "object") {
              semaphore++;

              setTimeout(
                function(a, b, c, d, callback) {
                  grabVariables(a, b, c, d, callback);
                },
                0,
                prop[p],
                newStr,
                level + 1,
                isDataLayerFlag || isDataLayer(p, prop),
                callback
              );
              //grabVariables(prop[p], newStr, level+1, isDataLayerFlag || isDataLayer(p, prop));
            } else if (isPrimitive(prop[p])) {
              //finally we reached a primitive value, that we would want to show to the user. Add it to the list.
              list.push({
                variableName: newStr,
                type: isDataLayerFlag ? "Data Layer" : "Global Variable",
                dataType: getType(prop[p]),
                value: prop[p]
              });
            }
          }
          if (isDataLayer(p, prop)) {
            dataLayers.push({
              variableName: newStr,
              type: "Data Layer",
              dataType: "Object"
            });
          }
        }
      }

      if (semaphore == 0) {
        callback.apply(this);
      }
    }

    const grabQueries = function() {
      var searchParams = new URLSearchParams(document.location.search);
      for (let p of searchParams) {
        list.push({
          variableName: p[0],
          type: "Query String",
          dataType: "String",
          value: p[1]
        });
      }
    };

    updateBlocklist();

    semaphore = 1;
    grabVariables(window, null, 0, false, function() {
      grabQueries();
      /*OTTInfra.pm({
          target: this._grid.contentWindow,
          type: "pageVariablesResult",
          data: {
            fullList: list,
            dataLayers: dataLayers
          }
      });*/
      //send it all back to the plugin / content.js
      postMessage(
        {
          key: "OTT",
          command: "pageVariablesResult",
          data: {
            fullList: list,
            dataLayers: dataLayers
          }
        },
        "*"
      );
    });
  };
  if (document.body) {
    getVarsFromPage();
  }
})();
