/**
 The MIT License

 Copyright (c) 2010 Daniel Park (http://metaweb.com, http://postmessage.freebaseapps.com)

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 **/

(function (window, $, undefined) {

    if (!$) {
        // jQuery is optional
        $ = {
            fn: {},
            extend: function () {
                var a = arguments[0];
                for (var i = 1, len = arguments.length; i < len; i++) {
                    var b = arguments[i];
                    for (var prop in b) {
                        a[prop] = b[prop];
                    }
                }
                return a;
            }
        };
    }

    $.fn.pm = function () {
        return this;
    };

    // send postmessage
    $.pm = function (options) {
        pm.send(options);
    };

    // bind postmessage handler
    $.pm.bind = function (type, fn, bindObj, origin, hash) {
        pm.bind(type, fn, bindObj, origin, hash);
    };

    // unbind postmessage handler
    $.pm.unbind = function (type, fn) {
        pm.unbind(type, fn);
    };

    // trigger a mimic post message event
    $.pm.dispatch = function (type, data, source, target) {
        pm._dispatch({
            data: JSON.stringify({ data: data, type: type }), source: source, target: target, origin: null
        });
    };

    // default postmessage origin on bind
    $.pm.origin = null;

    var pm = {

        send: function (options) {
            var o = $.extend({}, pm.defaults, options),
             target = o.target;
            if (!o.target) {
                return;
            }
            if (!o.type) {
                return;
            }
            var msg = { data: o.data, type: o.type };
            if (o.success) {
                msg.callback = pm._callback(o.success);
            }
            if (o.error) {
                msg.errback = pm._callback(o.error);
            }
            if (("postMessage" in target) && !o.hash) {
                pm._bind();
                target.postMessage(JSON.stringify(msg), o.origin || '*');
            }

        },

        bind: function (type, fn, bindObj, win, origin, hash) {
            if (("postMessage" in window) && !hash) {
                pm._bind(win);
            }
            else {
                return;
            }
            var l = pm.data("listeners.postmessage");
            if (!l) {
                l = {};
                pm.data("listeners.postmessage", l);
            }
            var fns = l[type];
            if (!fns) {
                fns = [];
                l[type] = fns;
            }
            fns.push({ fn: fn, origin: origin || $.pm.origin, bindObj: bindObj });
        },

        unbind: function (type, fn) {
            var l = pm.data("listeners.postmessage");
            if (l) {
                if (type) {
                    if (fn) {
                        // remove specific listener
                        var fns = l[type];
                        if (fns) {
                            var m = [];
                            for (var i = 0, len = fns.length; i < len; i++) {
                                var o = fns[i];
                                if (o.fn !== fn) {
                                    m.push(o);
                                }
                            }
                            l[type] = m;
                        }
                    }
                    else {
                        // remove all listeners by type
                        delete l[type];
                    }
                }
                else {
                    // unbind all listeners of all type
                    l = {};
                }
            }
        },

        data: function (k, v) {
            if (v === undefined) {
                return pm._data[k];
            }
            pm._data[k] = v;
            return v;
        },

        _data: {},

        _CHARS: '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.split(''),

        _random: function () {
            var r = [];
            for (var i = 0; i < 32; i++) {
                r[i] = pm._CHARS[0 | Math.random() * 32];
            }
            return r.join("");
        },

        _callback: function (fn) {
            var cbs = pm.data("callbacks.postmessage");
            if (!cbs) {
                cbs = {};
                pm.data("callbacks.postmessage", cbs);
            }
            var r = pm._random();
            cbs[r] = fn;
            return r;
        },

        _bind: function (win) {
            // are we already listening to message events on this w?
            if (!pm.data("listening.postmessage")) {
                win = win ? win : window;
                if (win.addEventListener) {
                    win.addEventListener("message", pm._dispatch, false);
                }
                else if (window.attachEvent) {
                    win.attachEvent("onmessage", pm._dispatch);
                }
                pm.data("listening.postmessage", 1);
            }
        },

        _dispatch: function (e) {
            var msg = {};
            try {
                msg = JSON.parse(e.data);
            }
            catch (ex) {
                return;
            }
            if (!msg.type) {
                return;
            }
            var cbs = pm.data("callbacks.postmessage") || {},
             cb = cbs[msg.type];
            if (cb) {
                cb(msg.data);
            }
            else {
                var l = pm.data("listeners.postmessage") || {};
                var fns = l[msg.type] || [];
                for (var i = 0, len = fns.length; i < len; i++) {
                    var o = fns[i];
                    if (o.origin && e.origin !== o.origin) {
                        if (msg.errback) {
                            // notify post message errback
                            var error = {
                                message: "postmessage origin mismatch",
                                origin: [e.origin, o.origin]
                            };
                            pm.send({ target: e.source, data: error, type: msg.errback });
                        }
                        continue;
                    }
                    try {
                        var r;
                        if (o.bindObj) {
                            r = o.fn.apply(o.bindObj, [msg.data, e]);
                        }
                        else {
                            r = o.fn(msg.data, e);
                        }
                        if (msg.callback) {
                            pm.send({ target: e.source, data: r, type: msg.callback });
                        }
                    }
                    catch (ex) {
                        if (msg.errback) {
                            // notify post message errback
                            pm.send({ target: e.source, data: ex, type: msg.errback });
                        }
                    }
                }
            }
        }
    };


    $.extend(pm, {
        defaults: {
            target: null,  /* target window (required) */
            url: null,     /* target window url (required if no window.postMessage or hash == true) */
            type: null,    /* message type (required) */
            data: null,    /* message data (required) */
            success: null, /* success callback (optional) */
            error: null,   /* error callback (optional) */
            origin: "*",   /* postmessage origin (optional) */
            hash: false    /* use location hash for message passing (optional) */
        }
    });

})(this, typeof (jQuery) === "undefined" ? null : jQuery);

